""" Triggers for events."""

from pydantic import BaseModel
from inductiva.client.models.trigger_task_type import TriggerTaskType


class Trigger(BaseModel):
    """Base class for triggers."""

    def get_trigger(self) -> None:
        raise NotImplementedError("Subclasses must implement this method.")


class TaskOutputUploaded(Trigger):
    """
    Trigger that is activated when a task output is uploaded.

    Attributes:
        task_id (int): ID of the task to monitor.
    """
    task_id: str

    def get_trigger(self):
        return {
            "task_id": self.task_id,
            "trigger": TriggerTaskType.TASK_OUTPUT_UPLOADED,
            "trigger_type": "task"
        }


class ObserverFileExists(Trigger):
    """
    Trigger that is activated when a file exists.
    
    Attributes:
        task_id (str): ID of the task to monitor.
        file_path (str): Path of the file to monitor."""
    task_id: str
    file_path: str

    def get_trigger(self):
        return {
            "trigger_type": "observer",
            "task_id": self.task_id,
            "trigger": "file_exists_observer",
            "file_path": self.file_path,
        }


class ObserverFileRegex(Trigger):
    """
    Trigger that is activated when a file matches a regex pattern.
    
    Attributes:
        task_id (str): ID of the task to monitor.
        file_path (str): Path of the file to monitor.
        regex (str): Regex pattern to match."""
    task_id: str
    file_path: str
    regex: str

    def get_trigger(self):
        return {
            "trigger_type": "observer",
            "task_id": self.task_id,
            "trigger": "file_regex_observer",
            "file_path": self.file_path,
            "regex": self.regex,
        }



"""List the storage contents via CLI."""

import argparse

import inductiva.client
import inductiva
from inductiva.utils import format_utils


def list_buckets(_):
    """List the user's buckets accessible via the Inductiva API."""
    api = inductiva.client.StorageApi(inductiva.api.get_client())
    response_list_buckets = api.list_buckets()
    response_default_bucket = api.get_default_bucket()

    buckets = [schema.model_dump() for schema in response_list_buckets]
    default_bucket = response_default_bucket.model_dump()
    for bucket in buckets:
        bucket["is_default"] = "yes" if bucket == default_bucket else "no"
        bucket["is_internal"] = "yes" if bucket["is_internal"] else "no"

    cols = list(buckets[0].keys())
    rows = [list(bucket.values()) for bucket in buckets]

    ansi_formatter = format_utils.get_ansi_formatter()
    headers = [lambda x: ansi_formatter(x.upper(), format_utils.Emphasis.BOLD)]

    table = format_utils.get_tabular_str(rows, cols, header_formatters=headers)
    print(table)


def register(parser):
    """
    Register the `inductiva storage bucket list` command for listing
    user buckets.
    """

    subparser = parser.add_parser(
        "list",
        aliases=["ls"],
        help="List all your storage buckets accessible via the Inductiva API.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva storage bucket list` command retrieves and displays "
        "your remote storage buckets available through the Inductiva API.\n"
        "This includes information such as the bucket name, region, provider, "
        "and whether it is internally managed by Inductiva.\n"
        "Example:\n"
        "  inductiva storage bucket list\n")

    subparser.set_defaults(func=list_buckets)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import json
import pprint
from pydantic import BaseModel, ConfigDict, Field, StrictStr, ValidationError, field_validator
from typing import Any, List, Optional
from inductiva.client.models.trigger_machine_group_create import TriggerMachineGroupCreate
from inductiva.client.models.trigger_observer_create import TriggerObserverCreate
from inductiva.client.models.trigger_task_create import TriggerTaskCreate
from pydantic import StrictStr, Field
from typing import Union, List, Set, Optional, Dict
from typing_extensions import Literal, Self

TRIGGER_ONE_OF_SCHEMAS = [
    "TriggerMachineGroupCreate", "TriggerObserverCreate", "TriggerTaskCreate"
]


class Trigger(BaseModel):
    """
    Trigger
    """
    # data type: TriggerTaskCreate
    oneof_schema_1_validator: Optional[TriggerTaskCreate] = None
    # data type: TriggerMachineGroupCreate
    oneof_schema_2_validator: Optional[TriggerMachineGroupCreate] = None
    # data type: TriggerObserverCreate
    oneof_schema_3_validator: Optional[TriggerObserverCreate] = None
    actual_instance: Optional[Union[TriggerMachineGroupCreate,
                                    TriggerObserverCreate,
                                    TriggerTaskCreate]] = None
    one_of_schemas: Set[str] = {
        "TriggerMachineGroupCreate", "TriggerObserverCreate",
        "TriggerTaskCreate"
    }

    model_config = ConfigDict(
        validate_assignment=True,
        protected_namespaces=(),
    )

    discriminator_value_class_map: Dict[str, str] = {}

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError(
                    "If a position argument is used, only 1 is allowed to set `actual_instance`"
                )
            if kwargs:
                raise ValueError(
                    "If a position argument is used, keyword arguments cannot be used."
                )
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator('actual_instance')
    def actual_instance_must_validate_oneof(cls, v):
        instance = Trigger.model_construct()
        error_messages = []
        match = 0
        # validate data type: TriggerTaskCreate
        if not isinstance(v, TriggerTaskCreate):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `TriggerTaskCreate`")
        else:
            match += 1
        # validate data type: TriggerMachineGroupCreate
        if not isinstance(v, TriggerMachineGroupCreate):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `TriggerMachineGroupCreate`"
            )
        else:
            match += 1
        # validate data type: TriggerObserverCreate
        if not isinstance(v, TriggerObserverCreate):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `TriggerObserverCreate`")
        else:
            match += 1
        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when setting `actual_instance` in Trigger with oneOf schemas: TriggerMachineGroupCreate, TriggerObserverCreate, TriggerTaskCreate. Details: "
                + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError(
                "No match found when setting `actual_instance` in Trigger with oneOf schemas: TriggerMachineGroupCreate, TriggerObserverCreate, TriggerTaskCreate. Details: "
                + ", ".join(error_messages))
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Union[str, Dict[str, Any]]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        match = 0

        # deserialize data into TriggerTaskCreate
        try:
            instance.actual_instance = TriggerTaskCreate.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into TriggerMachineGroupCreate
        try:
            instance.actual_instance = TriggerMachineGroupCreate.from_json(
                json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # deserialize data into TriggerObserverCreate
        try:
            instance.actual_instance = TriggerObserverCreate.from_json(json_str)
            match += 1
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if match > 1:
            # more than 1 match
            raise ValueError(
                "Multiple matches found when deserializing the JSON string into Trigger with oneOf schemas: TriggerMachineGroupCreate, TriggerObserverCreate, TriggerTaskCreate. Details: "
                + ", ".join(error_messages))
        elif match == 0:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into Trigger with oneOf schemas: TriggerMachineGroupCreate, TriggerObserverCreate, TriggerTaskCreate. Details: "
                + ", ".join(error_messages))
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(
                self.actual_instance.to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(
        self
    ) -> Optional[Union[Dict[str, Any], TriggerMachineGroupCreate,
                        TriggerObserverCreate, TriggerTaskCreate]]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(
                self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        else:
            # primitive type
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())



"""
Converts a Docker image to an Apptainer .sif file via CLI.
"""

import argparse
import threading
import sys
import os
import tempfile
from typing import TextIO
import uuid

from inductiva import constants

try:
    import docker
    from docker.errors import DockerException, ImageNotFound, NotFound
except ImportError:
    docker = None


def generator_thread(container, fout: TextIO):
    """Stream logs from a container."""
    for line in container.logs(stream=True):
        print(line.decode("utf-8"), end="", file=fout)


def join_container_streams(*containers, fout: TextIO = sys.stdout):
    """Join multiple container log streams into one output."""
    container_threads = []
    for container in containers:
        thread = threading.Thread(target=generator_thread,
                                  args=(container, fout))
        thread.start()
        container_threads.append(thread)
    for thread in container_threads:
        thread.join()


def convert_image(args, fout: TextIO = sys.stdout):
    """
    Converts a Docker image to an Apptainer .sif file.

    The source image may be provided as:
      - a Docker Hub URL (e.g. docker://python:3.11-slim),
      - a local Docker image reference (by image name or ID),
      - or a path to a tar archive.

    The output file will be saved to the specified location.
    """
    if docker is None:
        print(
            "Docker Python API not installed. Please run "
            "'pip install \"inductiva[task-runner]\"' to install it.",
            file=fout,
        )
        return False

    try:
        client = docker.from_env()
    except DockerException as e:
        print(f"Failed to connect to Docker: {e}", file=fout)
        return False

    # Prepare output file path
    output = os.path.abspath(args.output)
    outdir = os.path.dirname(output)
    if outdir and not os.path.exists(outdir):
        os.makedirs(outdir, exist_ok=True)
    os.chmod(outdir, 0o777)

    source = args.image
    conversion_source = None
    tmp_tar_path = None

    # Determine the conversion source
    if source.startswith("docker://"):
        # Use the docker:// URL directly.
        conversion_source = source
    elif os.path.exists(source):
        # Assume it's a tar file.
        conversion_source = f"docker-archive:///{os.path.abspath(source)}"
    else:
        # Assume it's a local Docker image reference.
        try:
            image_obj = client.images.get(source)
        except ImageNotFound:
            print(
                f"Local image '{source}' not found. "
                "\nThe image must be pulled or built in the system",
                file=fout)
            return False

        # Save the image to a temporary tar file.
        with tempfile.NamedTemporaryFile(suffix=".tar",
                                         delete=False,
                                         dir=constants.TMP_DIR) as tmp:
            tmp_tar_path = tmp.name
            os.chmod(tmp_tar_path, 0o644)
            with open(tmp_tar_path, "wb") as f:
                for chunk in image_obj.save(named=True):
                    f.write(chunk)
        conversion_source = f"docker-archive:///{tmp_tar_path}"

    mounts = []

    # If .tar or local docker image
    if tmp_tar_path:
        tar_host_path = os.path.abspath(tmp_tar_path)
        tar_host_dir = os.path.dirname(tar_host_path)
        tar_filename = os.path.basename(tar_host_path)
        tar_inside_container = f"/input/{tar_filename}"
        conversion_source = f"docker-archive://{tar_inside_container}"
        mounts.append(
            docker.types.Mount(target="/input",
                               source=tar_host_dir,
                               type="bind"))

    output_filename = os.path.basename(output)
    out_mount_path = "/output"
    mounts.append(
        docker.types.Mount(target=out_mount_path,
                           source=os.path.abspath(outdir),
                           type="bind"))

    command = [
        "build", f"{out_mount_path}/{output_filename}", conversion_source
    ]

    container = None
    try:
        container = client.containers.run(
            image=constants.TASK_RUNNER_IMAGE,
            name=f"apptainer-converter-{uuid.uuid4().hex[:8]}",
            entrypoint="/usr/bin/apptainer",
            command=command,
            mounts=mounts,
            platform="linux/amd64",
            detach=True,
            auto_remove=False,
        )

        print("Conversion started with container ID:",
              container.short_id,
              file=fout)
        join_container_streams(container, fout=fout)
        print(f"Conversion complete. Output saved at: {output}", file=fout)

    except KeyboardInterrupt:
        print("\n⚠️ Conversion interrupted by user. Cleaning up...", file=fout)
        if container:
            try:
                container.kill()
                container.remove(force=True)
                print("🧹 Container stopped and removed.", file=fout)
            except Exception as cleanup_err:  # pylint: disable=broad-exception-caught
                print(f"⚠️ Failed to clean up container: {cleanup_err}",
                      file=fout)
        return False

    except Exception as e:  # pylint: disable=broad-exception-caught
        print(f"Error during conversion: {e}", file=fout)
        return False

    finally:
        # Clean up the container if it was created.
        try:
            if container:
                container.remove(force=True)
        except NotFound:
            pass

    # Clean up the temporary tar file if one was created.
    if tmp_tar_path and os.path.exists(tmp_tar_path):
        try:
            os.remove(tmp_tar_path)
        except Exception as e:  # pylint: disable=broad-exception-caught
            print(
                f"Warning: could not remove temporary file {tmp_tar_path}: {e}",
                file=fout)
            return False

    return True


def register(parser):
    """Register the convert command."""
    subparser = parser.add_parser(
        "convert",
        help="Convert a Docker image to an Apptainer .sif file.",
        formatter_class=argparse.RawTextHelpFormatter,
    )
    subparser.description = (
        "The `inductiva containers convert` command converts a Docker "
        "image—specified as a Docker Hub URL,\n a local Docker image reference,"
        " or a .tar archive—into an Apptainer .sif file using\n the Apptainer.")
    subparser.add_argument(
        "image",
        type=str,
        help=("Docker image reference. A docker:// URL, a local image name/ID, "
              "or a path to a .tar file."),
    )
    subparser.add_argument(
        "output",
        type=str,
        help="Path to save the resulting .sif file.",
    )
    subparser.set_defaults(func=convert_image)



"""Top command for task."""
from typing import TextIO
import argparse
import sys

from inductiva import tasks


def top(args: argparse.Namespace, fout: TextIO = sys.stdout):
    """Prints the result of the `top -b -H -n 1` command.
    
    This command will list the processes and threads (-H) in batch mode (-b).
    This command will run only once (-n 1) instead of running continuously.
    The result is an instant snapshot of the machine CPU and RAM metrics.
    """
    task_id = args.id
    task = tasks.Task(task_id)
    # pylint: disable=protected-access
    result, return_code = task._top()
    if result:
        print(result, file=fout)
    return return_code


def register(parser):
    """Register the `top` command for monitoring task resource usage."""
    subparser = parser.add_parser(
        "top",
        help="Displays the output of the `top` command from the machine running"
        " the task.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva tasks top` command streams the output of the"
        "`top -b -H -n 1` command executed on the machine where the "
        "task is running. This allows real-time monitoring of system "
        "processes and resource usage for the task.")

    subparser.add_argument("id",
                           type=str,
                           help="The ID of the task to monitor.")

    subparser.set_defaults(func=top)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
from inspect import getfullargspec
import json
import pprint
import re  # noqa: F401
from pydantic import BaseModel, ConfigDict, Field, StrictStr, ValidationError, field_validator
from typing import Optional
from inductiva.client.models.action_email_info import ActionEmailInfo
from inductiva.client.models.action_webhook_info import ActionWebhookInfo
from typing import Union, Any, List, Set, TYPE_CHECKING, Optional, Dict
from typing_extensions import Literal, Self
from pydantic import Field

ACTION1_ANY_OF_SCHEMAS = ["ActionEmailInfo", "ActionWebhookInfo"]


class Action1(BaseModel):
    """
    Action1
    """

    # data type: ActionEmailInfo
    anyof_schema_1_validator: Optional[ActionEmailInfo] = None
    # data type: ActionWebhookInfo
    anyof_schema_2_validator: Optional[ActionWebhookInfo] = None
    if TYPE_CHECKING:
        actual_instance: Optional[Union[ActionEmailInfo,
                                        ActionWebhookInfo]] = None
    else:
        actual_instance: Any = None
    any_of_schemas: Set[str] = {"ActionEmailInfo", "ActionWebhookInfo"}

    model_config = {
        "validate_assignment": True,
        "protected_namespaces": (),
    }

    def __init__(self, *args, **kwargs) -> None:
        if args:
            if len(args) > 1:
                raise ValueError(
                    "If a position argument is used, only 1 is allowed to set `actual_instance`"
                )
            if kwargs:
                raise ValueError(
                    "If a position argument is used, keyword arguments cannot be used."
                )
            super().__init__(actual_instance=args[0])
        else:
            super().__init__(**kwargs)

    @field_validator('actual_instance')
    def actual_instance_must_validate_anyof(cls, v):
        instance = Action1.model_construct()
        error_messages = []
        # validate data type: ActionEmailInfo
        if not isinstance(v, ActionEmailInfo):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `ActionEmailInfo`")
        else:
            return v

        # validate data type: ActionWebhookInfo
        if not isinstance(v, ActionWebhookInfo):
            error_messages.append(
                f"Error! Input type `{type(v)}` is not `ActionWebhookInfo`")
        else:
            return v

        if error_messages:
            # no match
            raise ValueError(
                "No match found when setting the actual_instance in Action1 with anyOf schemas: ActionEmailInfo, ActionWebhookInfo. Details: "
                + ", ".join(error_messages))
        else:
            return v

    @classmethod
    def from_dict(cls, obj: Dict[str, Any]) -> Self:
        return cls.from_json(json.dumps(obj))

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Returns the object represented by the json string"""
        instance = cls.model_construct()
        error_messages = []
        # anyof_schema_1_validator: Optional[ActionEmailInfo] = None
        try:
            instance.actual_instance = ActionEmailInfo.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))
        # anyof_schema_2_validator: Optional[ActionWebhookInfo] = None
        try:
            instance.actual_instance = ActionWebhookInfo.from_json(json_str)
            return instance
        except (ValidationError, ValueError) as e:
            error_messages.append(str(e))

        if error_messages:
            # no match
            raise ValueError(
                "No match found when deserializing the JSON string into Action1 with anyOf schemas: ActionEmailInfo, ActionWebhookInfo. Details: "
                + ", ".join(error_messages))
        else:
            return instance

    def to_json(self) -> str:
        """Returns the JSON representation of the actual instance"""
        if self.actual_instance is None:
            return "null"

        if hasattr(self.actual_instance, "to_json") and callable(
                self.actual_instance.to_json):
            return self.actual_instance.to_json()
        else:
            return json.dumps(self.actual_instance)

    def to_dict(
        self
    ) -> Optional[Union[Dict[str, Any], ActionEmailInfo, ActionWebhookInfo]]:
        """Returns the dict representation of the actual instance"""
        if self.actual_instance is None:
            return None

        if hasattr(self.actual_instance, "to_dict") and callable(
                self.actual_instance.to_dict):
            return self.actual_instance.to_dict()
        else:
            return self.actual_instance

    def to_str(self) -> str:
        """Returns the string representation of the actual instance"""
        return pprint.pformat(self.model_dump())



"""Inductiva CLI.

`main` is the entrypoint for the CLI. It creates a parser and registers
subparsers for several subcommands (e.g. `tasks` and `machines`).
"""
from . import utils
from .main import main, get_main_parser



"""Register CLI commands for quotas."""
import argparse
import os

from inductiva import constants
from inductiva._cli import loader, utils


def register(root_parser):
    parser = root_parser.add_parser(
        "quotas",
        help="Quotas management utilities.",
        formatter_class=argparse.RawTextHelpFormatter)

    parser.description = ("Quotas management utilities.\n\n"
                          "The `inductiva quotas` command allows you to "
                          "consult your user's internal quotas.\n")

    utils.show_help_msg(parser)

    subparsers = parser.add_subparsers(title="available subcomands")
    loader.load_commands(subparsers,
                         os.path.dirname(__file__),
                         package=__name__,
                         ignores_prefix=constants.LOADER_IGNORE_PREFIX,
                         hides_prefix=constants.LOADER_HIDE_PREFIX)



#pylint: disable=missing-module-docstring
from inductiva.events.event import (list_events, remove, get, register)
from inductiva.events.triggers import Trigger, TaskOutputUploaded
from inductiva.events.actions import Action, EmailNotification



"""Register CLI command for logout."""
import argparse
import os

from inductiva import constants


def logout(_):
    """
    Logs the user out by removing the stored API key.

    The function will remove the stored API key, logging the user out.
    """
    if not os.path.exists(constants.API_KEY_FILE_PATH):
        print("Error: You are not logged in.")
        return

    os.remove(constants.API_KEY_FILE_PATH)
    print("Logout successful.")


def register(parser):
    """Register the logout command."""
    subparser = parser.add_parser(
        "logout",
        help="Logout by removing the Inductiva API key.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "The `inductiva logout` command allows you to log out by removing "
        "the stored API key.\n")

    subparser.set_defaults(func=logout)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, StrictBool, StrictFloat, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional, Union
from inductiva.client.models.autoscale_policy import AutoscalePolicy
from inductiva.client.models.cpu_info import CPUInfo
from inductiva.client.models.dynamic_disk_resize_config import DynamicDiskResizeConfig
from inductiva.client.models.gpu_info import GPUInfo
from inductiva.client.models.machine_group_cost_per_hour import MachineGroupCostPerHour
from inductiva.client.models.machine_group_costs import MachineGroupCosts
from inductiva.client.models.machine_group_status import MachineGroupStatus
from inductiva.client.models.machine_group_type import MachineGroupType
from inductiva.client.models.machine_group_usage_statistics import MachineGroupUsageStatistics
from inductiva.client.models.machine_info import MachineInfo
from inductiva.client.models.providers import Providers
from inductiva.client.models.sharing_level import SharingLevel
from typing import Optional, Set
from typing_extensions import Self


class VMGroupConfig(BaseModel):
    """
    Schema for creating an instance group.
    """

  # noqa: E501
    max_idle_time: Optional[Union[StrictFloat, StrictInt]] = None
    auto_terminate_ts: Optional[datetime] = None
    id: Optional[StrictStr] = None
    name: Optional[StrictStr] = None
    machine_type: Optional[StrictStr] = None
    disk_size_gb: Optional[StrictInt] = None
    total_ram_gb: Optional[StrictInt] = None
    threads_per_core: Optional[StrictInt] = None
    num_vms: Optional[StrictInt] = None
    idle_seconds: Optional[StrictInt] = None
    creation_timestamp: Optional[datetime] = None
    deletion_timestamp: Optional[datetime] = None
    type: Optional[MachineGroupType] = None
    provider_id: Optional[Providers] = None
    started: Optional[StrictBool] = False
    quota_usage: Optional[Dict[str, Union[StrictFloat, StrictInt]]] = None
    status: Optional[MachineGroupStatus] = None
    spot: Optional[StrictBool] = False
    is_elastic: Optional[StrictBool] = False
    min_vms: Optional[StrictInt] = None
    max_vms: Optional[StrictInt] = None
    autoscale_policy: Optional[AutoscalePolicy] = None
    dynamic_disk_resize_config: Optional[DynamicDiskResizeConfig] = None
    custom_vm_image: Optional[StrictStr] = None
    machines: Optional[List[MachineInfo]] = None
    provider_description: Optional[StrictStr] = None
    costs: Optional[MachineGroupCosts] = None
    cost_per_hour: Optional[MachineGroupCostPerHour] = None
    statistics: Optional[MachineGroupUsageStatistics] = None
    sharing_level: Optional[SharingLevel] = None
    zone: StrictStr
    gpu_info: Optional[GPUInfo] = None
    cpu_info: Optional[CPUInfo] = None
    __properties: ClassVar[List[str]] = [
        "max_idle_time", "auto_terminate_ts", "id", "name", "machine_type",
        "disk_size_gb", "total_ram_gb", "threads_per_core", "num_vms",
        "idle_seconds", "creation_timestamp", "deletion_timestamp", "type",
        "provider_id", "started", "quota_usage", "status", "spot", "is_elastic",
        "min_vms", "max_vms", "autoscale_policy", "dynamic_disk_resize_config",
        "custom_vm_image", "machines", "provider_description", "costs",
        "cost_per_hour", "statistics", "sharing_level", "zone", "gpu_info",
        "cpu_info"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of VMGroupConfig from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # override the default output from pydantic by calling `to_dict()` of dynamic_disk_resize_config
        if self.dynamic_disk_resize_config:
            _dict[
                'dynamic_disk_resize_config'] = self.dynamic_disk_resize_config.to_dict(
                )
        # override the default output from pydantic by calling `to_dict()` of each item in machines (list)
        _items = []
        if self.machines:
            for _item_machines in self.machines:
                if _item_machines:
                    _items.append(_item_machines.to_dict())
            _dict['machines'] = _items
        # override the default output from pydantic by calling `to_dict()` of costs
        if self.costs:
            _dict['costs'] = self.costs.to_dict()
        # override the default output from pydantic by calling `to_dict()` of cost_per_hour
        if self.cost_per_hour:
            _dict['cost_per_hour'] = self.cost_per_hour.to_dict()
        # override the default output from pydantic by calling `to_dict()` of statistics
        if self.statistics:
            _dict['statistics'] = self.statistics.to_dict()
        # override the default output from pydantic by calling `to_dict()` of gpu_info
        if self.gpu_info:
            _dict['gpu_info'] = self.gpu_info.to_dict()
        # override the default output from pydantic by calling `to_dict()` of cpu_info
        if self.cpu_info:
            _dict['cpu_info'] = self.cpu_info.to_dict()
        # set to None if max_idle_time (nullable) is None
        # and model_fields_set contains the field
        if self.max_idle_time is None and "max_idle_time" in self.model_fields_set:
            _dict['max_idle_time'] = None

        # set to None if auto_terminate_ts (nullable) is None
        # and model_fields_set contains the field
        if self.auto_terminate_ts is None and "auto_terminate_ts" in self.model_fields_set:
            _dict['auto_terminate_ts'] = None

        # set to None if id (nullable) is None
        # and model_fields_set contains the field
        if self.id is None and "id" in self.model_fields_set:
            _dict['id'] = None

        # set to None if name (nullable) is None
        # and model_fields_set contains the field
        if self.name is None and "name" in self.model_fields_set:
            _dict['name'] = None

        # set to None if machine_type (nullable) is None
        # and model_fields_set contains the field
        if self.machine_type is None and "machine_type" in self.model_fields_set:
            _dict['machine_type'] = None

        # set to None if disk_size_gb (nullable) is None
        # and model_fields_set contains the field
        if self.disk_size_gb is None and "disk_size_gb" in self.model_fields_set:
            _dict['disk_size_gb'] = None

        # set to None if total_ram_gb (nullable) is None
        # and model_fields_set contains the field
        if self.total_ram_gb is None and "total_ram_gb" in self.model_fields_set:
            _dict['total_ram_gb'] = None

        # set to None if threads_per_core (nullable) is None
        # and model_fields_set contains the field
        if self.threads_per_core is None and "threads_per_core" in self.model_fields_set:
            _dict['threads_per_core'] = None

        # set to None if num_vms (nullable) is None
        # and model_fields_set contains the field
        if self.num_vms is None and "num_vms" in self.model_fields_set:
            _dict['num_vms'] = None

        # set to None if idle_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.idle_seconds is None and "idle_seconds" in self.model_fields_set:
            _dict['idle_seconds'] = None

        # set to None if creation_timestamp (nullable) is None
        # and model_fields_set contains the field
        if self.creation_timestamp is None and "creation_timestamp" in self.model_fields_set:
            _dict['creation_timestamp'] = None

        # set to None if deletion_timestamp (nullable) is None
        # and model_fields_set contains the field
        if self.deletion_timestamp is None and "deletion_timestamp" in self.model_fields_set:
            _dict['deletion_timestamp'] = None

        # set to None if quota_usage (nullable) is None
        # and model_fields_set contains the field
        if self.quota_usage is None and "quota_usage" in self.model_fields_set:
            _dict['quota_usage'] = None

        # set to None if status (nullable) is None
        # and model_fields_set contains the field
        if self.status is None and "status" in self.model_fields_set:
            _dict['status'] = None

        # set to None if min_vms (nullable) is None
        # and model_fields_set contains the field
        if self.min_vms is None and "min_vms" in self.model_fields_set:
            _dict['min_vms'] = None

        # set to None if max_vms (nullable) is None
        # and model_fields_set contains the field
        if self.max_vms is None and "max_vms" in self.model_fields_set:
            _dict['max_vms'] = None

        # set to None if dynamic_disk_resize_config (nullable) is None
        # and model_fields_set contains the field
        if self.dynamic_disk_resize_config is None and "dynamic_disk_resize_config" in self.model_fields_set:
            _dict['dynamic_disk_resize_config'] = None

        # set to None if custom_vm_image (nullable) is None
        # and model_fields_set contains the field
        if self.custom_vm_image is None and "custom_vm_image" in self.model_fields_set:
            _dict['custom_vm_image'] = None

        # set to None if provider_description (nullable) is None
        # and model_fields_set contains the field
        if self.provider_description is None and "provider_description" in self.model_fields_set:
            _dict['provider_description'] = None

        # set to None if costs (nullable) is None
        # and model_fields_set contains the field
        if self.costs is None and "costs" in self.model_fields_set:
            _dict['costs'] = None

        # set to None if cost_per_hour (nullable) is None
        # and model_fields_set contains the field
        if self.cost_per_hour is None and "cost_per_hour" in self.model_fields_set:
            _dict['cost_per_hour'] = None

        # set to None if statistics (nullable) is None
        # and model_fields_set contains the field
        if self.statistics is None and "statistics" in self.model_fields_set:
            _dict['statistics'] = None

        # set to None if sharing_level (nullable) is None
        # and model_fields_set contains the field
        if self.sharing_level is None and "sharing_level" in self.model_fields_set:
            _dict['sharing_level'] = None

        # set to None if gpu_info (nullable) is None
        # and model_fields_set contains the field
        if self.gpu_info is None and "gpu_info" in self.model_fields_set:
            _dict['gpu_info'] = None

        # set to None if cpu_info (nullable) is None
        # and model_fields_set contains the field
        if self.cpu_info is None and "cpu_info" in self.model_fields_set:
            _dict['cpu_info'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of VMGroupConfig from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "max_idle_time":
                obj.get("max_idle_time"),
            "auto_terminate_ts":
                obj.get("auto_terminate_ts"),
            "id":
                obj.get("id"),
            "name":
                obj.get("name"),
            "machine_type":
                obj.get("machine_type"),
            "disk_size_gb":
                obj.get("disk_size_gb"),
            "total_ram_gb":
                obj.get("total_ram_gb"),
            "threads_per_core":
                obj.get("threads_per_core"),
            "num_vms":
                obj.get("num_vms"),
            "idle_seconds":
                obj.get("idle_seconds"),
            "creation_timestamp":
                obj.get("creation_timestamp"),
            "deletion_timestamp":
                obj.get("deletion_timestamp"),
            "type":
                obj.get("type"),
            "provider_id":
                obj.get("provider_id"),
            "started":
                obj.get("started") if obj.get("started") is not None else False,
            "quota_usage":
                obj.get("quota_usage"),
            "status":
                obj.get("status"),
            "spot":
                obj.get("spot") if obj.get("spot") is not None else False,
            "is_elastic":
                obj.get("is_elastic")
                if obj.get("is_elastic") is not None else False,
            "min_vms":
                obj.get("min_vms"),
            "max_vms":
                obj.get("max_vms"),
            "autoscale_policy":
                obj.get("autoscale_policy"),
            "dynamic_disk_resize_config":
                DynamicDiskResizeConfig.from_dict(
                    obj["dynamic_disk_resize_config"])
                if obj.get("dynamic_disk_resize_config") is not None else None,
            "custom_vm_image":
                obj.get("custom_vm_image"),
            "machines": [
                MachineInfo.from_dict(_item) for _item in obj["machines"]
            ] if obj.get("machines") is not None else None,
            "provider_description":
                obj.get("provider_description"),
            "costs":
                MachineGroupCosts.from_dict(obj["costs"])
                if obj.get("costs") is not None else None,
            "cost_per_hour":
                MachineGroupCostPerHour.from_dict(obj["cost_per_hour"])
                if obj.get("cost_per_hour") is not None else None,
            "statistics":
                MachineGroupUsageStatistics.from_dict(obj["statistics"])
                if obj.get("statistics") is not None else None,
            "sharing_level":
                obj.get("sharing_level"),
            "zone":
                obj.get("zone"),
            "gpu_info":
                GPUInfo.from_dict(obj["gpu_info"])
                if obj.get("gpu_info") is not None else None,
            "cpu_info":
                CPUInfo.from_dict(obj["cpu_info"])
                if obj.get("cpu_info") is not None else None
        })
        return _obj



"""CLI commands to terminate computational resources."""
import logging
import sys

import argparse

from inductiva import resources
import inductiva
from inductiva.utils import input_functions
from ...localization import translator as __


def terminate_machine_group(args):
    """Terminate one or all computational resources."""
    names = args.name
    all_names = args.all
    confirm = args.confirm

    if names and all_names:
        print(
            "inductiva resources terminate: error: "
            "argument name not allowed with argument --all",
            file=sys.stderr)
        return 1

    active_machines = resources.get()

    if not active_machines:
        print("No active resources to terminate.")
        return 0

    if not all_names and not names:
        print("No resource(s) specified.\n"
              "> Use `inductiva resources terminate -h` for help.")
        return 1

    # dict to map from name to machine
    name_to_machine = {machine.name: machine for machine in active_machines}
    # the user can give the same name multiple times
    target_machine_names = set(names or name_to_machine.keys())
    invalid_names = target_machine_names.difference(name_to_machine)

    if invalid_names:
        for name in invalid_names:
            print(f"Resource {name} does not exist.")
        print("Aborting.")
        return 1

    confirm = confirm or input_functions.user_confirmation_prompt(
        target_machine_names, __("resources-prompt-terminate-all"),
        __("resources-prompt-terminate-big", len(target_machine_names)),
        __("resources-prompt-terminate-small"), False)

    if not confirm:
        return 0

    print_quotas = False

    before_quotas = inductiva.users.get_quotas()
    for name in target_machine_names:
        name_to_machine[name].terminate(verbose=False)
        logging.info("Successfully requested termination of %s.",
                     name_to_machine[name].name)
        if (name_to_machine[name].provider ==
                inductiva.resources.utils.ProviderType.GCP):
            print_quotas = True

    if print_quotas:
        cls = inductiva.resources.MachineGroup
        base_machine = cls.__new__(cls)
        rows = ["max_price_hour", "max_vcpus", "max_instances"]
        base_machine.quota_usage = {k: before_quotas[k].in_use for k in rows}
        logging.info(base_machine.quota_usage_table_str("freed by resources"))

    return 0


def register(parser):
    """Register the terminate command for the resources."""

    subparser = parser.add_parser("terminate",
                                  help="Terminate resources.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva resources terminate` command "
                             "provides a utility for terminating\n"
                             "active computational resources. It allows you"
                             " to specify the names of the resources\n"
                             "to terminate, or terminate all active resources."
                             " Multiple resources can be terminated\n"
                             "at once by providing their names.\n\n")

    subparser.add_argument("name",
                           type=str,
                           help="Name(s) of the resource(s) to terminate.",
                           nargs="*")
    subparser.add_argument("-y",
                           "--yes",
                           action="store_true",
                           dest="confirm",
                           default=False,
                           help="Sets any confirmation values to \"yes\" "
                           "automatically. Users will not be asked for "
                           "confirmation to terminate resource(s).")
    subparser.add_argument("--all",
                           action="store_true",
                           help="Terminate all machines.")

    subparser.set_defaults(func=terminate_machine_group)



"""SWASH module of the API."""
from pathlib import Path
from typing import Optional, Union

from inductiva import simulators, tasks, types
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class SWASH(simulators.Simulator):
    """Class to invoke a generic SWASH simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the SWASH simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "swash"

    def run(self,
            input_dir: Optional[str],
            sim_config_filename: str,
            *,
            command: str = "swashrun",
            use_hwthread: bool = True,
            n_vcpus: Optional[int] = None,
            storage_dir: Optional[str] = "",
            on: types.ComputationalResources,
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            sim_config_filename: Name of the simulation configuration file.
            on: The computational resource to launch the simulation on.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            command: The command to run the simulation. Default is 'swashrun'.
                The user can also specify 'swash.exe'.
            storage_dir: Directory for storing simulation results.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
        """

        if command not in ("swashrun", "swash.exe"):
            raise ValueError("Invalid command. Use 'swashrun' or 'swash.exe'.")

        if sim_config_filename is None and command == "swashrun":
            raise ValueError("Simulation configuration file "
                             "(sim_config_filename) not provided.\n"
                             "When using 'swashrun' it is mandatory to provide "
                             "sim_config_filename.")

        commands = []

        path_config_filename = Path(sim_config_filename)

        if path_config_filename.is_absolute():
            raise ValueError("sim_config_filename must be a path relative to "
                             "the input directory.")

        self._input_files_exist(input_dir=input_dir,
                                remote_assets=remote_assets,
                                sim_config_filename=sim_config_filename)

        working_dir = path_config_filename.parent
        config_file_only = path_config_filename.name

        # swashrun uses internal MPI
        # we call apptainer run ... swashrun ... -mpi np
        if command == "swashrun":

            machinefile_command = Command(
                "dd if=/dev/stdin of=machinefile",
                f"localhost slots={on.available_vcpus}")

            commands.append(machinefile_command)

            #if the user does not provide n_vcpus use all available by default
            mpi_flag = f"-mpi {n_vcpus or on.available_vcpus}"

            swashrun_command = Command(
                f"swashrun -input {config_file_only} {mpi_flag}")
            commands.append(swashrun_command)

        # we call mpirun ... apptainer ... swash.exe
        # works with clusters
        elif command == "swash.exe":

            mpi_kwargs = {}
            #If the user does not provide n_vcpus mpi will use all available
            if n_vcpus is not None:
                mpi_kwargs["np"] = n_vcpus
            mpi_kwargs["use_hwthread_cpus"] = use_hwthread

            mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)
            swash_exe_command = Command(f"swash.exe {sim_config_filename}",
                                        mpi_config=mpi_config)
            commands.append(swash_exe_command)

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           run_subprocess_dir=str(working_dir),
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""Base class for low-level simulators."""
import logging
import os
import pathlib
import re
from abc import ABC
from typing import Literal, Optional, Union

from inductiva import commands, logs, projects, resources, tasks, types

from .methods import list_available_images


def mpi_enabled(cls):
    """Class decorator that adds MPICluster to the supported resources.
    """
    supports = getattr(cls, "_supported_resources", set())
    cls._supported_resources = supports | {resources.MPICluster}  # pylint: disable=protected-access

    return cls


class Simulator(ABC):
    """Base simulator class."""

    _supported_resources = {
        resources.MachineGroup, resources.ElasticMachineGroup
    }
    _logger = logging.getLogger(__name__)

    _supported_device_values = ["gpu", "cpu", "auto"]

    def __init__(self,
                 /,
                 version: Optional[str] = None,
                 use_dev: bool = False,
                 device: Literal["auto", "cpu", "gpu"] = "auto"):
        """Initialize the simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
            device (str): Specifies whether to use the CPU or GPU version of
                the Docker image. If `auto` is picked we will pick the
                appropriate device based on the machine used to run the
                simulation, meaning, if the machine has a GPU we will pick
                GPU, otherwise we pick CPU.
        """
        if version is not None and not isinstance(version, str):
            raise ValueError("Version must be a string or None.")

        if  device is not None and \
            device.lower() not in self._supported_device_values:
            raise ValueError("Wrong value for `device`. Supported "
                             "values are `gpu`, `cpu` or `auto`.")
        self.simulator = ""
        self.simulator_name_alias = None
        self._version = version
        self._use_dev = bool(use_dev)
        self._image_uri = self._get_image_uri()
        self.container_image = self._image_uri
        self._logger.info("")
        self._device = device.lower() if device else None

    @property
    def version(self):
        """Get the version of the simulator."""
        return self._version

    @property
    def name(self):
        """Get the name of the simulator."""
        return self.__class__.__name__

    @property
    def image_uri(self):
        """Get the image URI for this simulator."""
        return self._image_uri

    def _input_files_exist(self, input_dir, remote_assets, **kwargs):
        """
        Checks if all the files in kwargs are present in the input_dir.
        """
        if remote_assets is not None:
            return

        missing_files = []

        for _, file_path in kwargs.items():
            # Get the full file path by joining the input directory with the
            # file path in kwargs
            full_file_path = os.path.join(input_dir, file_path)

            # Check if the file exists
            if not os.path.isfile(full_file_path):
                missing_files.append(file_path)

        if missing_files:
            missing_files_str = ", ".join(missing_files)
            raise FileNotFoundError(
                "The following files are missing from your input directory:\n"
                f"{missing_files_str}")

    def _regex_exists_in_file(self, file_path: str, pattern: str) -> bool:
        """
        Check if a given regular expression exists in a file.

        :param file_path: Path to the file.
        :param pattern: Regular expression pattern to search for.
        :return: True if the pattern exists, False otherwise.
        """
        with open(file_path, "r", encoding="utf-8") as file:
            for line in file:
                if re.search(pattern, line):
                    return True
        return False

    def _check_vcpus(
        self,
        n_vcpus: Optional[int],
        machine: types.ComputationalResources,
    ) -> bool:
        """ Checks if the machine supports the number of n_vcpus passed."""
        if n_vcpus is not None and n_vcpus > machine.n_vcpus.total:
            raise ValueError(
                "The number of virtual cpus asked surpasses the"
                " available virtual cpus for the selected resource.")

    def _get_image_uri(self):
        """Get the appropriate image name for this simulator."""

        img_type = "development" if self._use_dev else "production"
        sim_name = self.name
        name = sim_name.lower()

        available = list_available_images()
        listing = available.get(img_type, {}).get(name, [])

        self._supported_versions = {
            element.split("_")[0] for element in listing
        }

        self._supported_versions_with_suffixes = listing

        if self._version is None:
            # kutu does not have a specific tag for the latest version
            # this hack is a workaround to get that version, but it is prone
            # to errors.
            self._version = max(self._supported_versions)

        if self._version not in self._supported_versions:
            raise ValueError(
                f"Version {self.version} is not available for simulator {name}."
                f" Available versions are: {self._supported_versions}.")
        self._logger.info("■ Using %s image of %s version %s", img_type,
                          sim_name, self.version)

        return f"docker://inductiva/kutu:{name}_v{self._version}"

    @classmethod
    def get_supported_resources(cls):
        """Get the supported computational resources for this simulator."""
        return tuple(cls._supported_resources)

    def _setup_input_dir(self, input_dir: str):
        """Setup the simulator input directory."""
        input_dir_path = pathlib.Path(input_dir)
        if not input_dir_path.is_dir():
            raise ValueError(
                f"The provided path (\"{input_dir}\") is not a directory.")
        return input_dir_path

    def _validate_input_files(self, input_dir, remote_assets):
        if input_dir == "":
            raise ValueError(
                "input_dir cannot be an empty string. Use None instead.")

        if not input_dir and not remote_assets:
            raise ValueError(
                "Either input_dir or remote_assets must be provided.")

    def _get_version_suffixes(self,
                              resource: types.ComputationalResources) -> str:
        """
        Gets the suffixes based on the resource used for the simulation.

        If the resource has a GPU we will prioritize that.

        :param resource: The resource used for the simulation.
        :return: A string of suffixes.
        """
        dev_suffix = "_dev" if self._use_dev else ""
        gpu_suffix = "_gpu" if resource.has_gpu() and (
            f"{self.version}_gpu"
            in self._supported_versions_with_suffixes) else ""

        # Overwrites the gpu suffix if the user passed a specific
        # device
        if self._device == "gpu":
            gpu_suffix = "_gpu"
        elif self._device == "cpu":
            gpu_suffix = ""

        # Add new suffixes here
        suffix = f"{gpu_suffix}"

        if self._device and self._device == "gpu" and not resource.has_gpu():
            raise ValueError("The selected machine needs to have a GPU.\n"
                             "If available, either pick device=`cpu` or pick"
                             " a different machine.")

        suffix = f"{suffix}{dev_suffix}"
        return suffix

    def _append_simulator_image_suffix_based_on_resource(
            self, image_uri: str, resource: types.ComputationalResources):
        """
        Get the simulator image for the simulation, based on the resourced used.

        This method takes in consideration the specified device for the
        simulation. If device=`auto` we will pick the appropriate device based
        on the machine.

        :param resource: The computational resource to use for the simulation.
        :return: The simulator image based on the resource used.
        """
        suffixes = ""

        if self._device:
            suffixes = self._get_version_suffixes(resource)

        return f"{image_uri}{suffixes}"

    @logs.mute_logging()
    def run(
        self,
        input_dir: Optional[str],
        *_args,
        on: types.ComputationalResources,
        storage_dir: Optional[str] = "",
        resubmit_on_preemption: bool = False,
        remote_assets: Optional[Union[str, list[str]]] = None,
        project: Optional[str] = None,
        time_to_live: Optional[str] = None,
        **kwargs,
    ) -> tasks.Task:
        """Run the simulation.

        Args:
            on: The computational resource to launch the simulation in. If None
                the simulation is launched in a machine of the default pool.
            input_dir: Path to the directory containing the input files.
            _args: Unused in this method, but defined to allow for more
                non-default arguments in method override in subclasses.
            storage_dir: Parent directory for storing simulation
                               results.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
            **kwargs: Additional keyword arguments to be passed to the
                simulation API method.
        """
        if isinstance(remote_assets, str):
            remote_assets = [remote_assets]

        self._validate_input_files(input_dir, remote_assets)

        input_dir_path = self._setup_input_dir(input_dir) if input_dir else None

        if on is None:
            raise ValueError(
                "A default computational resource is no longer "
                "provided. Please specify a computational resource. "
                "Check https://docs.inductiva.ai/en/latest/how_to/"
                "run-parallel_simulations.html "
                "to learn how to create your own computational resource.")

        self._validate_computational_resources(on)

        if "commands" in kwargs:
            cmds = commands.Command.commands_to_dicts(kwargs["commands"])
            kwargs["commands"] = cmds

        # Get the user-specified image name. If not specified,
        # use the default image name for the current simulator
        image_uri = kwargs.pop("container_image", self._image_uri)

        # CustomImage does not use suffixes. We can the image as is
        if self.__class__.__name__ != "CustomImage":
            image_uri = self._append_simulator_image_suffix_based_on_resource(
                image_uri, on)

        if on.has_gpu() and "_gpu" not in image_uri:
            logging.warning("Attention: The machine you selected has a GPU, but"
                            " the simulator you picked will run on the CPU "
                            "only.\n")

        # This will create the project if it doesn't exist
        if project:
            projects.Project(project)

        return tasks.run_simulation(
            self.simulator,
            input_dir_path,
            simulator_obj=self,
            storage_dir=storage_dir,
            machine_group=on,
            container_image=image_uri,
            resubmit_on_preemption=resubmit_on_preemption,
            remote_assets=remote_assets,
            simulator_name_alias=self.simulator_name_alias,
            project_name=project,
            time_to_live=time_to_live,
            **kwargs,
        )

    def _validate_computational_resources(self, resource):
        """Validate the computational resources passed to the run method.

        Args:
            resource: The computational resource to validate.
            valid_resources: The valid computational resources for the simulator
        """

        if not isinstance(resource, tuple(self._supported_resources)):
            raise ValueError(
                "The computational resource is invalid for this simulator. "
                f"Expected one of {self._supported_resources} but got "
                f"{resource}.")



"""Constants that can be set for the Inductiva client."""
import os
import pathlib
import platform

TURN_SERVER_URL = os.environ.get("INDUCTIVA_TURN_SERVER_URL",
                                 "webrtc.inductiva.ai:3478")

TASK_RUNNER_IMAGE = os.environ.get("INDUCTIVA_TASK_RUNNER_IMAGE",
                                   "inductiva/task-runner:main")

FILE_TRACKER_IMAGE = os.environ.get("INDUCTIVA_FILE_TRACKER_IMAGE",
                                    "inductiva/file-tracker:main")

APPTAINER_CONVERTER_IMAGE = os.environ.get(
    "INDUCTIVA_APPTAINER_CONVERTER_IMAGE",
    "inductiva/kutu:apptainer-converter_v0.1.0_dev")

TASK_KILL_MAX_API_REQUESTS = 5

TASK_KILL_RETRY_SLEEP_SEC = 1.0

MAX_CONFIRMATION_LINES = 5

LOADER_COMMAND_PREFIX = "cmd_"

LOADER_IGNORE_PREFIX = "__"

LOADER_HIDE_PREFIX = "_"

INDUCTIVA_GIT_EXAMPLES_URL = ("https://raw.githubusercontent.com/inductiva"
                              "/inductiva/refs/heads/main/inductiva/tests/"
                              "test_simulators/")

# when printing the stack trace, how many lines to show
EXCEPTIONS_MAX_TRACEBACK_DEPTH = 2

TASK_FAILED_LINES_TO_DUMP = 10

INDUCTIVA_LOGS_WAIT_SLEEP_TIME = 1.0

system = platform.system()
if platform.system().lower() == "windows":
    HOME_DIR = pathlib.Path.home() / "AppData" / "inductiva"
elif platform.system().lower() in ["linux", "darwin"]:
    HOME_DIR = pathlib.Path.home() / ".inductiva"
else:
    raise RuntimeError(f"Current operating system {system} not supported.")

LOGS_FILE_PATH = HOME_DIR / "inductiva.log"
API_KEY_FILE_PATH = HOME_DIR / "api_key"
TMP_DIR = HOME_DIR / "tmp"
TMP_DIR.mkdir(parents=True, exist_ok=True)

TASK_OUTPUT_ZIP = "output.zip"
TASK_INPUT_ZIP = "input.zip"

LOGIN_MESSAGE = "Please login with `inductiva auth login`."



"""CaNS module of the API for numerical simulations of fluid flows."""
from typing import Literal, Optional, Union

from inductiva import simulators, tasks, types
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class CaNS(simulators.Simulator):
    """Class to invoke a generic CaNS simulation on the API.

    """

    def __init__(self,
                 /,
                 version: Optional[str] = None,
                 use_dev: bool = False,
                 device: Literal["auto", "cpu", "gpu"] = "auto"):
        """Initialize the CaNS simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
            device (str): Select between CPU or GPU for running the simulation.
                Default is "auto", which will auto-detect if the machine has a
                GPU and use if it available, otherwise use the CPU.
        """
        super().__init__(version=version, use_dev=use_dev, device=device)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "cans"

    def run(self,
            input_dir: Optional[str],
            sim_config_filename: str,
            *,
            on: types.ComputationalResources,
            use_hwthread: bool = True,
            n_vcpus: Optional[int] = None,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
            other arguments: See the documentation of the base class.
        """

        self._input_files_exist(input_dir=input_dir,
                                remote_assets=remote_assets,
                                sim_config_filename=sim_config_filename)

        #To run cans on gpu the n_vcpus must be equal to the number of gpus
        if on.has_gpu and n_vcpus is None:
            n_vcpus = on.gpu_count()

        mpi_kwargs = {}
        mpi_kwargs["use_hwthread_cpus"] = use_hwthread
        if n_vcpus is not None:
            mpi_kwargs["np"] = n_vcpus

        mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)
        commands = [
            "mkdir -p data",
            Command(f"cans {sim_config_filename}", mpi_config=mpi_config)
        ]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



# coding: utf-8

# flake8: noqa
"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

# import models into model package
from inductiva.client.models.action import Action
from inductiva.client.models.action1 import Action1
from inductiva.client.models.action_email_create import ActionEmailCreate
from inductiva.client.models.action_email_info import ActionEmailInfo
from inductiva.client.models.action_webhook_create import ActionWebhookCreate
from inductiva.client.models.action_webhook_info import ActionWebhookInfo
from inductiva.client.models.autoscale_policy import AutoscalePolicy
from inductiva.client.models.available_simulator_versions import AvailableSimulatorVersions
from inductiva.client.models.backend_version import BackendVersion
from inductiva.client.models.bucket import Bucket
from inductiva.client.models.bulk_simulator_operation import BulkSimulatorOperation
from inductiva.client.models.bulk_simulator_response import BulkSimulatorResponse
from inductiva.client.models.cpu_info import CPUInfo
from inductiva.client.models.capability import Capability
from inductiva.client.models.compression_method import CompressionMethod
from inductiva.client.models.cost_components import CostComponents
from inductiva.client.models.cost_detail import CostDetail
from inductiva.client.models.cost_type import CostType
from inductiva.client.models.currency_code import CurrencyCode
from inductiva.client.models.disk_resize_request import DiskResizeRequest
from inductiva.client.models.dynamic_disk_resize_config import DynamicDiskResizeConfig
from inductiva.client.models.event_create import EventCreate
from inductiva.client.models.event_info import EventInfo
from inductiva.client.models.executer import Executer
from inductiva.client.models.file_info import FileInfo
from inductiva.client.models.gpu_info import GPUInfo
from inductiva.client.models.http_validation_error import HTTPValidationError
from inductiva.client.models.machine_group_cost_per_hour import MachineGroupCostPerHour
from inductiva.client.models.machine_group_costs import MachineGroupCosts
from inductiva.client.models.machine_group_status import MachineGroupStatus
from inductiva.client.models.machine_group_time_stat import MachineGroupTimeStat
from inductiva.client.models.machine_group_type import MachineGroupType
from inductiva.client.models.machine_group_usage_statistics import MachineGroupUsageStatistics
from inductiva.client.models.machine_group_with_costs import MachineGroupWithCosts
from inductiva.client.models.machine_info import MachineInfo
from inductiva.client.models.machine_operation import MachineOperation
from inductiva.client.models.machine_operation_type import MachineOperationType
from inductiva.client.models.machine_type import MachineType
from inductiva.client.models.multi_part_export_operation import MultiPartExportOperation
from inductiva.client.models.operation_end import OperationEnd
from inductiva.client.models.operation_start import OperationStart
from inductiva.client.models.operation_status import OperationStatus
from inductiva.client.models.operation_type import OperationType
from inductiva.client.models.order import Order
from inductiva.client.models.organization_cost_components import OrganizationCostComponents
from inductiva.client.models.organization_costs import OrganizationCosts
from inductiva.client.models.processor_type import ProcessorType
from inductiva.client.models.project import Project
from inductiva.client.models.project_create import ProjectCreate
from inductiva.client.models.project_statistics import ProjectStatistics
from inductiva.client.models.project_type import ProjectType
from inductiva.client.models.project_update import ProjectUpdate
from inductiva.client.models.providers import Providers
from inductiva.client.models.quota import Quota
from inductiva.client.models.quota_scope import QuotaScope
from inductiva.client.models.register_vm_group_request import RegisterVMGroupRequest
from inductiva.client.models.select_mode import SelectMode
from inductiva.client.models.sharing_level import SharingLevel
from inductiva.client.models.simulator_create import SimulatorCreate
from inductiva.client.models.simulator_response import SimulatorResponse
from inductiva.client.models.simulator_update import SimulatorUpdate
from inductiva.client.models.simulator_version import SimulatorVersion
from inductiva.client.models.simulator_with_machine_types import SimulatorWithMachineTypes
from inductiva.client.models.spot import Spot
from inductiva.client.models.spot1 import Spot1
from inductiva.client.models.storage_contents_with_pagination import StorageContentsWithPagination
from inductiva.client.models.storage_cost import StorageCost
from inductiva.client.models.storage_file_info import StorageFileInfo
from inductiva.client.models.storage_operation import StorageOperation
from inductiva.client.models.storage_operation_name import StorageOperationName
from inductiva.client.models.storage_sort_by import StorageSortBy
from inductiva.client.models.task import Task
from inductiva.client.models.task_connection import TaskConnection
from inductiva.client.models.task_machine_operation import TaskMachineOperation
from inductiva.client.models.task_metric_create import TaskMetricCreate
from inductiva.client.models.task_metrics import TaskMetrics
from inductiva.client.models.task_operation import TaskOperation
from inductiva.client.models.task_operation_name import TaskOperationName
from inductiva.client.models.task_position_in_queue import TaskPositionInQueue
from inductiva.client.models.task_request import TaskRequest
from inductiva.client.models.task_runner_api_connection_info import TaskRunnerAPIConnectionInfo
from inductiva.client.models.task_runner_register_info import TaskRunnerRegisterInfo
from inductiva.client.models.task_status import TaskStatus
from inductiva.client.models.task_status_code import TaskStatusCode
from inductiva.client.models.task_status_info import TaskStatusInfo
from inductiva.client.models.task_step import TaskStep
from inductiva.client.models.task_submitted_info import TaskSubmittedInfo
from inductiva.client.models.task_with_pagination import TaskWithPagination
from inductiva.client.models.task_with_status_history import TaskWithStatusHistory
from inductiva.client.models.terms_and_conditions import TermsAndConditions
from inductiva.client.models.top_up_type import TopUpType
from inductiva.client.models.transaction import Transaction
from inductiva.client.models.trigger import Trigger
from inductiva.client.models.trigger1 import Trigger1
from inductiva.client.models.trigger_credits_info import TriggerCreditsInfo
from inductiva.client.models.trigger_credits_type import TriggerCreditsType
from inductiva.client.models.trigger_file_exists_observer_info import TriggerFileExistsObserverInfo
from inductiva.client.models.trigger_machine_group_create import TriggerMachineGroupCreate
from inductiva.client.models.trigger_machine_group_info import TriggerMachineGroupInfo
from inductiva.client.models.trigger_machine_group_type import TriggerMachineGroupType
from inductiva.client.models.trigger_observer_create import TriggerObserverCreate
from inductiva.client.models.trigger_observer_type import TriggerObserverType
from inductiva.client.models.trigger_regex_observer_info import TriggerRegexObserverInfo
from inductiva.client.models.trigger_task_create import TriggerTaskCreate
from inductiva.client.models.trigger_task_info import TriggerTaskInfo
from inductiva.client.models.trigger_task_type import TriggerTaskType
from inductiva.client.models.trigger_type import TriggerType
from inductiva.client.models.update_operation_status import UpdateOperationStatus
from inductiva.client.models.upload_part import UploadPart
from inductiva.client.models.usage_statistics import UsageStatistics
from inductiva.client.models.user import User
from inductiva.client.models.user_activity import UserActivity
from inductiva.client.models.user_computation_trend import UserComputationTrend
from inductiva.client.models.user_costs import UserCosts
from inductiva.client.models.user_costs_details import UserCostsDetails
from inductiva.client.models.user_costs_over_time import UserCostsOverTime
from inductiva.client.models.user_most_used_machine_types_overview import UserMostUsedMachineTypesOverview
from inductiva.client.models.user_most_used_simulators_overview import UserMostUsedSimulatorsOverview
from inductiva.client.models.user_task_status_overview import UserTaskStatusOverview
from inductiva.client.models.user_tasks_overview import UserTasksOverview
from inductiva.client.models.vm_group_config import VMGroupConfig
from inductiva.client.models.validation_error import ValidationError
from inductiva.client.models.validation_error_loc_inner import ValidationErrorLocInner
from inductiva.client.models.version_comparison_result import VersionComparisonResult
from inductiva.client.models.zip_archive_info import ZipArchiveInfo
from inductiva.client.models.zip_file_range import ZipFileRange



"""FDS simulator module of the API."""

from typing import Optional, Union
import logging

from inductiva import simulators, tasks, types
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


class FDS(simulators.Simulator):
    """Class to invoke a generic FDS simulation on the API."""

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the FDS simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "fds"

    def run(self,
            input_dir: Optional[str],
            sim_config_filename: str,
            *,
            on: types.ComputationalResources,
            n_vcpus: Optional[int] = None,
            n_omp_threads: Optional[int] = None,
            n_mpi_processes: Optional[int] = None,
            use_hwthread: bool = True,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            sim_config_filename: Name of the simulation configuration file.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            n_omp_threads: Number of OpenMP threads to use in the simulation.
                If not provided, it defaults to the number of available vcpus
                of the machine group where the task will run.
            n_mpi_processes: Number of MPI processes that will run the
                simulation. If not provided, it defaults to 1. Note that the
                number of MPI processes can't exceed the number of meshes
                of the simulation case.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            other arguments: See the documentation of the base class.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
        """

        self._input_files_exist(input_dir=input_dir,
                                remote_assets=remote_assets,
                                sim_config_filename=sim_config_filename)

        available_mpi_slots = on.get_available_mpi_slots(
            use_hwthread=use_hwthread)

        available_vcpus = on.available_vcpus

        if n_vcpus is None:
            logging.info(
                "Param n_vcpus not set. Defaulting to the number of "
                "available vcpus (%s).\n", available_vcpus)
            n_vcpus = available_vcpus

        if n_mpi_processes is None:
            n_mpi_processes = 1

            logging.info(
                "Param n_mpi_processes not set. Defaulting to %s. "
                "Note that the number of MPI processes for FDS "
                "simulations is limited by the number of meshes "
                "in the simulation case.\n", n_mpi_processes)

        if n_omp_threads is None:
            n_omp_threads = max(n_vcpus // n_mpi_processes, 1)
            logging.info("Param n_omp_threads not set. Defaulting to %s.\n",
                         n_omp_threads)

        if n_mpi_processes > available_mpi_slots:
            raise ValueError(
                f"n_mpi_processes ({n_mpi_processes}) exceeds the number of "
                f"MPI slots available on the machine ({available_mpi_slots})")

        if n_vcpus == 0:
            raise ValueError("n_vcpus must be larger than 0")
        if n_mpi_processes == 0:
            raise ValueError("n_mpi_processes must be larger than 0")
        if n_omp_threads == 0:
            raise ValueError("n_omp_threads must be larger than 0")

        requested_vcpus = n_omp_threads * n_mpi_processes
        if requested_vcpus > available_vcpus:
            raise ValueError(
                f"n_mpi_processes * n_omp_threads ({n_mpi_processes} * "
                f"{n_omp_threads} = {requested_vcpus}) can't be larger than "
                f"{available_vcpus} (the number of available VCPUs in "
                "the specified machine)")

        mpi_config = None
        if n_mpi_processes > 1:
            mpi_kwargs = {}
            mpi_kwargs["use_hwthread_cpus"] = use_hwthread
            mpi_kwargs["np"] = n_mpi_processes
            mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)

        commands = [
            Command(
                "/opt/fds/Build/ompi_gnu_linux/fds_ompi_gnu_linux "
                f"{sim_config_filename}",
                mpi_config=mpi_config,
                env={"OMP_NUM_THREADS": str(n_omp_threads)},
            )
        ]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



# flake8: noqa

# import apis into api package
from inductiva.client.api.compute_api import ComputeApi
from inductiva.client.api.events_api import EventsApi
from inductiva.client.api.projects_api import ProjectsApi
from inductiva.client.api.simulators_api import SimulatorsApi
from inductiva.client.api.storage_api import StorageApi
from inductiva.client.api.tasks_api import TasksApi
from inductiva.client.api.users_api import UsersApi
from inductiva.client.api.version_api import VersionApi



"""FVCOM simulator module of the API."""
from typing import Optional, Union

from inductiva import simulators, tasks, types
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class FVCOM(simulators.Simulator):
    """Class to invoke a generic FVCOM simulation on the API.

    """

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the FVCOM simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "fvcom"

    def run(self,
            input_dir: Optional[str],
            *,
            on: types.ComputationalResources,
            debug: int = 0,
            model: str = "",
            case_name: str = "",
            use_hwthread: bool = True,
            create_namelist: str = "",
            n_vcpus: Optional[int] = None,
            working_dir: Optional[str] = "",
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            casename: Name of the simulation case.

            on: The computational resource to launch the simulation on. If None
                the simulation is submitted to a machine in the default pool.

            debug: Debug level of the simulation (from 0 to 7).

            model: At the current moment we provide users with two
            options:
                - None (default): Uses default fvcom binary.
                - 'estuary': Uses the fvcom_estuary binary.
                The modules used to compile each binary can be found in the
                kutu repository, in the make.inc and make_estuary.inc files
                https://github.com/inductiva/kutu/tree/main/simulators/fvcom/

            create_namelist: Used to create a namelist file for the simulation.
                Example: 'create_namelist=hello' will create hello_run.nml in
                the working_dir.

            working_dir: Path (relative to the input directory) to the directory
                where the simulation nml file is located. If not provided, the
                input directory is used.

            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.

            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.

            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiates with
                `spot=True`.

            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.

            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.

            other arguments: See the documentation of the base class.
        """
        if model != "" and model.lower() != "estuary":
            raise ValueError(
                f"Invalid model: {model}. Valid options are None or 'estuary'.")

        if case_name:
            case_name = f"--CASENAME={case_name}"

        if create_namelist:
            create_namelist = f"--CREATE_NAMELIST {create_namelist}"

        if model:
            model = f"_{model.lower()}"

        cmd = f"fvcom{model} {case_name} {create_namelist} --dbg={debug}"
        mpi_kwargs = {}
        mpi_kwargs["use_hwthread_cpus"] = use_hwthread
        if n_vcpus is not None:
            mpi_kwargs["np"] = n_vcpus

        mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)
        commands = [Command(cmd, mpi_config=mpi_config)]

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           run_subprocess_dir=working_dir,
                           resubmit_on_preemption=resubmit_on_preemption,
                           remote_assets=remote_assets,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



"""CLI for logs."""
import sys

from .. import utils as cli_utils
from ... import tasks


def stream_task_logs_tail(args):
    task_id = args.mode.lower()
    task = tasks.Task(task_id)

    filename = "stdout.txt" if args.stdout else "stderr.txt"
    files = ["stdout.txt", "stderr.txt"
            ] if args.stdout == args.stderr else [filename]

    return task.tail_files(files, 10, True, sys.stdout)


def register(parser):
    cli_utils.show_help_msg(parser)

    parser.add_argument(
        "mode",
        type=str,
        nargs="?",
        default="SUBMITTED",
        help=
        ("Mode of log retrieval. "
         "Use 'SUBMITTED' for the last submitted task, "
         "'SUBMITTED-1' for the second last submitted task, and so on. "
         "'STARTED' and 'STARTED-n' follow the same pattern for started tasks. "
         "Or, use a specific 'task_id' to retrieve logs for a particular task."
        ))

    parser.add_argument("--stdout",
                        action="store_true",
                        help="Consumes the standard output stream of the task.")
    parser.add_argument("--stderr",
                        action="store_true",
                        help="Consumes the standard error stream of the task.")
    parser.add_argument("--no-color",
                        action="store_true",
                        help="Disables the colorized output.")
    parser.add_argument(
        "--wait",
        "-w",
        action="store_true",
        help=("Wait for the task to start running before consuming the logs.\n"
              "Without this flag, the logs will be consumed immediately\nor "
              "returns an error if the task is not running."))
    # Register function to call when this subcommand is used
    parser.set_defaults(func=stream_task_logs_tail)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from pydantic import BaseModel, ConfigDict, StrictFloat, StrictInt
from typing import Any, ClassVar, Dict, List, Optional, Union
from typing import Optional, Set
from typing_extensions import Self


class TaskMetrics(BaseModel):
    """
    TaskMetrics
    """

  # noqa: E501
    total_seconds: Optional[Union[StrictFloat, StrictInt]] = None
    container_image_download_seconds: Optional[Union[StrictFloat,
                                                     StrictInt]] = None
    container_image_size_bytes: Optional[StrictInt] = None
    queue_time_seconds: Optional[Union[StrictFloat, StrictInt]] = None
    computation_seconds: Optional[Union[StrictFloat, StrictInt]] = None
    input_upload_seconds: Optional[Union[StrictFloat, StrictInt]] = None
    input_download_seconds: Optional[Union[StrictFloat, StrictInt]] = None
    input_decompression_seconds: Optional[Union[StrictFloat, StrictInt]] = None
    output_upload_seconds: Optional[Union[StrictFloat, StrictInt]] = None
    input_zipped_size_bytes: Optional[StrictInt] = None
    input_size_bytes: Optional[StrictInt] = None
    output_total_files: Optional[StrictInt] = None
    output_size_bytes: Optional[StrictInt] = None
    output_zipped_size_bytes: Optional[StrictInt] = None
    output_compression_seconds: Optional[Union[StrictFloat, StrictInt]] = None
    __properties: ClassVar[List[str]] = [
        "total_seconds", "container_image_download_seconds",
        "container_image_size_bytes", "queue_time_seconds",
        "computation_seconds", "input_upload_seconds", "input_download_seconds",
        "input_decompression_seconds", "output_upload_seconds",
        "input_zipped_size_bytes", "input_size_bytes", "output_total_files",
        "output_size_bytes", "output_zipped_size_bytes",
        "output_compression_seconds"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of TaskMetrics from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        # set to None if total_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.total_seconds is None and "total_seconds" in self.model_fields_set:
            _dict['total_seconds'] = None

        # set to None if container_image_download_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.container_image_download_seconds is None and "container_image_download_seconds" in self.model_fields_set:
            _dict['container_image_download_seconds'] = None

        # set to None if container_image_size_bytes (nullable) is None
        # and model_fields_set contains the field
        if self.container_image_size_bytes is None and "container_image_size_bytes" in self.model_fields_set:
            _dict['container_image_size_bytes'] = None

        # set to None if queue_time_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.queue_time_seconds is None and "queue_time_seconds" in self.model_fields_set:
            _dict['queue_time_seconds'] = None

        # set to None if computation_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.computation_seconds is None and "computation_seconds" in self.model_fields_set:
            _dict['computation_seconds'] = None

        # set to None if input_upload_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.input_upload_seconds is None and "input_upload_seconds" in self.model_fields_set:
            _dict['input_upload_seconds'] = None

        # set to None if input_download_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.input_download_seconds is None and "input_download_seconds" in self.model_fields_set:
            _dict['input_download_seconds'] = None

        # set to None if input_decompression_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.input_decompression_seconds is None and "input_decompression_seconds" in self.model_fields_set:
            _dict['input_decompression_seconds'] = None

        # set to None if output_upload_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.output_upload_seconds is None and "output_upload_seconds" in self.model_fields_set:
            _dict['output_upload_seconds'] = None

        # set to None if input_zipped_size_bytes (nullable) is None
        # and model_fields_set contains the field
        if self.input_zipped_size_bytes is None and "input_zipped_size_bytes" in self.model_fields_set:
            _dict['input_zipped_size_bytes'] = None

        # set to None if input_size_bytes (nullable) is None
        # and model_fields_set contains the field
        if self.input_size_bytes is None and "input_size_bytes" in self.model_fields_set:
            _dict['input_size_bytes'] = None

        # set to None if output_total_files (nullable) is None
        # and model_fields_set contains the field
        if self.output_total_files is None and "output_total_files" in self.model_fields_set:
            _dict['output_total_files'] = None

        # set to None if output_size_bytes (nullable) is None
        # and model_fields_set contains the field
        if self.output_size_bytes is None and "output_size_bytes" in self.model_fields_set:
            _dict['output_size_bytes'] = None

        # set to None if output_zipped_size_bytes (nullable) is None
        # and model_fields_set contains the field
        if self.output_zipped_size_bytes is None and "output_zipped_size_bytes" in self.model_fields_set:
            _dict['output_zipped_size_bytes'] = None

        # set to None if output_compression_seconds (nullable) is None
        # and model_fields_set contains the field
        if self.output_compression_seconds is None and "output_compression_seconds" in self.model_fields_set:
            _dict['output_compression_seconds'] = None

        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of TaskMetrics from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "total_seconds":
                obj.get("total_seconds"),
            "container_image_download_seconds":
                obj.get("container_image_download_seconds"),
            "container_image_size_bytes":
                obj.get("container_image_size_bytes"),
            "queue_time_seconds":
                obj.get("queue_time_seconds"),
            "computation_seconds":
                obj.get("computation_seconds"),
            "input_upload_seconds":
                obj.get("input_upload_seconds"),
            "input_download_seconds":
                obj.get("input_download_seconds"),
            "input_decompression_seconds":
                obj.get("input_decompression_seconds"),
            "output_upload_seconds":
                obj.get("output_upload_seconds"),
            "input_zipped_size_bytes":
                obj.get("input_zipped_size_bytes"),
            "input_size_bytes":
                obj.get("input_size_bytes"),
            "output_total_files":
                obj.get("output_total_files"),
            "output_size_bytes":
                obj.get("output_size_bytes"),
            "output_zipped_size_bytes":
                obj.get("output_zipped_size_bytes"),
            "output_compression_seconds":
                obj.get("output_compression_seconds")
        })
        return _obj



"""Wrapper class for simulator commands."""
from typing import Optional, Dict
import re
from inductiva import types
from .mpiconfig import MPIConfig


class Command:
    """Abstraction class for commands."""
    SPECIAL_CHARACTERS = r"[|><&;*?~$]"
    SPECIAL_CHARACTERS_PRETTY_PRINT = r"| > < & ; * ? ~ $"

    def __init__(
        self,
        cmd: str,
        *prompts: str,
        mpi_config: Optional[MPIConfig] = None,
        env: Optional[Dict[str, str]] = None,
    ):
        """
        Args:
          cmd: A string with the command, e.g, 'gmx pdb2gmx -f protein.pdb'
          *promts: Strings to be used as answers to prompts, e.g, 'y'

        Example:
        >>> cmd = Command("gmx pdb2gmx -f protein.pdb", "y", "y")
        """
        if not isinstance(cmd, str):
            raise ValueError("cmd argument must be a string.")

        if not all(isinstance(prompt, str) for prompt in prompts):
            raise ValueError("Prompts must be all strings.")

        if mpi_config and not isinstance(mpi_config, MPIConfig):
            raise TypeError("'mpi_config' must be an instance of MPIConfig.")
        if self._has_special_chars(cmd):
            raise ValueError(
                "Command contains unsupported characters.\n"
                "Please avoid using any of the following characters:\n"
                f"{self.SPECIAL_CHARACTERS_PRETTY_PRINT}\n")
        self.cmd = cmd
        self.prompts = list(prompts)
        self.mpi_config = mpi_config
        self.env = env or {}

    def _has_special_chars(self, command):
        """Checks if the command contains special characters.
        Args:
            command: The command to check.
        Returns:
            True if the command contains special characters, False otherwise.
            """

        return bool(re.search(self.SPECIAL_CHARACTERS, command))

    def to_dict(self):
        """
        Convert this command to a dictionary representation.

        Returns:
        A dictionary with {"cmd": self.cmd, "prompts": self.prompts}

        Example:
        >>> cmd = Command("gmx pdb2gmx -f protein.pdb", "y", "y")
        >>> cmd.to_dict()
        >>> {"cmd": "gmx pdb2gmx -f protein.pdb", prompts: ["y", "y"]}
        """
        ret_dict = self.__dict__.copy()

        if ret_dict["mpi_config"] is not None:
            ret_dict["mpi_config"] = ret_dict["mpi_config"].to_dict()

        return ret_dict

    @staticmethod
    def commands_to_dicts(commands: types.Commands):
        """
        Convert the given commands to a list of dictionaries.

        Example:
        >>> commands = [Command("run", "y"), "terminate"]
        >>> simulator.commands_to_dict(commands)
        >>> [{"cmd": "run", "prompts": ["y"]},
        ...  {"cmd": "terminate", "prompts": []}]
        """
        return [
            cmd.to_dict()
            if isinstance(cmd, Command) else Command(cmd).to_dict()
            for cmd in commands
        ]



"""
List all container files inside a specified (or default) storage folder.
"""

import argparse
from inductiva import storage, client


def list_containers(args):
    default_folder = "my-containers"
    output_path = args.folder

    if not output_path:
        output_path = default_folder

    try:
        storage.listdir(output_path, args.max_results)
    except client.exceptions.ApiException as e:
        if e.status == 404:
            error_msg = f"Folder '{output_path}' not found on remote storage."
            if output_path == default_folder:
                error_msg += ("\nUse the `inductiva containers upload` command "
                              "to upload a container to the remote storage.")
            print(error_msg)
            return False
    except Exception:  # pylint: disable=broad-exception-caught
        print("Unkown error while listing containers.")
        return False


def register(parser):
    """Register the upload-container command."""
    subparser = parser.add_parser(
        "list",
        aliases=["ls"],
        help="List all container files in remote storage.",
        formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = (
        "Lists all container files in the folder within remote storage.\n"
        "If no folder is provided, defaults to the 'my-containers' directory.")

    subparser.add_argument(
        "folder",
        nargs="?",
        type=str,
        help=(
            "Opt folder path in remote storage to list container files from.\n"
            "If omitted, defaults to 'my-containers'."),
    )

    subparser.add_argument("-m", "--max-results", default=10, type=int)

    subparser.set_defaults(func=list_containers)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

import warnings
from pydantic import validate_call, Field, StrictFloat, StrictStr, StrictInt
from typing import Any, Dict, List, Optional, Tuple, Union
from typing_extensions import Annotated

from pydantic import StrictInt, StrictStr
from typing import Any, List, Optional
from inductiva.client.models.event_create import EventCreate
from inductiva.client.models.event_info import EventInfo
from inductiva.client.models.trigger_machine_group_type import TriggerMachineGroupType
from inductiva.client.models.trigger_observer_type import TriggerObserverType
from inductiva.client.models.trigger_task_type import TriggerTaskType
from inductiva.client.models.trigger_type import TriggerType

from inductiva.client.api_client import ApiClient, RequestSerialized
from inductiva.client.api_response import ApiResponse
from inductiva.client.rest import RESTResponseType


class EventsApi:
    """NOTE: This class is auto generated by OpenAPI Generator
    Ref: https://openapi-generator.tech

    Do not edit the class manually.
    """

    def __init__(self, api_client=None) -> None:
        if api_client is None:
            api_client = ApiClient.get_default()
        self.api_client = api_client

    @validate_call
    def create_event(
        self,
        event_create: EventCreate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> object:
        """Create Event

        Add a new event.

        :param event_create: (required)
        :type event_create: EventCreate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_event_serialize(event_create=event_create,
                                              _request_auth=_request_auth,
                                              _content_type=_content_type,
                                              _headers=_headers,
                                              _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def create_event_with_http_info(
        self,
        event_create: EventCreate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[object]:
        """Create Event

        Add a new event.

        :param event_create: (required)
        :type event_create: EventCreate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_event_serialize(event_create=event_create,
                                              _request_auth=_request_auth,
                                              _content_type=_content_type,
                                              _headers=_headers,
                                              _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def create_event_without_preload_content(
        self,
        event_create: EventCreate,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Create Event

        Add a new event.

        :param event_create: (required)
        :type event_create: EventCreate
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._create_event_serialize(event_create=event_create,
                                              _request_auth=_request_auth,
                                              _content_type=_content_type,
                                              _headers=_headers,
                                              _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '201': "object",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _create_event_serialize(
        self,
        event_create,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter
        if event_create is not None:
            _body_params = event_create

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # set the HTTP header `Content-Type`
        if _content_type:
            _header_params['Content-Type'] = _content_type
        else:
            _default_content_type = (self.api_client.select_header_content_type(
                ['application/json']))
            if _default_content_type is not None:
                _header_params['Content-Type'] = _default_content_type

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='POST',
            resource_path='/events/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def delete_event(
        self,
        event_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> None:
        """Delete Event

        Delete a event by its ID.

        :param event_id: (required)
        :type event_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_event_serialize(event_id=event_id,
                                              _request_auth=_request_auth,
                                              _content_type=_content_type,
                                              _headers=_headers,
                                              _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def delete_event_with_http_info(
        self,
        event_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[None]:
        """Delete Event

        Delete a event by its ID.

        :param event_id: (required)
        :type event_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_event_serialize(event_id=event_id,
                                              _request_auth=_request_auth,
                                              _content_type=_content_type,
                                              _headers=_headers,
                                              _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def delete_event_without_preload_content(
        self,
        event_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Delete Event

        Delete a event by its ID.

        :param event_id: (required)
        :type event_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._delete_event_serialize(event_id=event_id,
                                              _request_auth=_request_auth,
                                              _content_type=_content_type,
                                              _headers=_headers,
                                              _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '204': None,
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _delete_event_serialize(
        self,
        event_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if event_id is not None:
            _path_params['event_id'] = event_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='DELETE',
            resource_path='/events/{event_id}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def get_event(
        self,
        event_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> EventInfo:
        """Get Event

        Get a event by its ID.

        :param event_id: (required)
        :type event_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_event_serialize(event_id=event_id,
                                           _request_auth=_request_auth,
                                           _content_type=_content_type,
                                           _headers=_headers,
                                           _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "EventInfo",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def get_event_with_http_info(
        self,
        event_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[EventInfo]:
        """Get Event

        Get a event by its ID.

        :param event_id: (required)
        :type event_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_event_serialize(event_id=event_id,
                                           _request_auth=_request_auth,
                                           _content_type=_content_type,
                                           _headers=_headers,
                                           _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "EventInfo",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def get_event_without_preload_content(
        self,
        event_id: StrictStr,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get Event

        Get a event by its ID.

        :param event_id: (required)
        :type event_id: str
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_event_serialize(event_id=event_id,
                                           _request_auth=_request_auth,
                                           _content_type=_content_type,
                                           _headers=_headers,
                                           _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "EventInfo",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _get_event_serialize(
        self,
        event_id,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        if event_id is not None:
            _path_params['event_id'] = event_id
        # process the query parameters
        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/events/{event_id}',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)

    @validate_call
    def get_events(
        self,
        trigger_type: Optional[TriggerType] = None,
        task_id: Optional[StrictStr] = None,
        task_trigger: Optional[TriggerTaskType] = None,
        machine_group_id: Optional[StrictInt] = None,
        machine_group_trigger: Optional[TriggerMachineGroupType] = None,
        observer_id: Optional[StrictStr] = None,
        observer_trigger: Optional[TriggerObserverType] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> List[EventInfo]:
        """Get Events

        Fetch events based on filters.

        :param trigger_type:
        :type trigger_type: TriggerType
        :param task_id:
        :type task_id: str
        :param task_trigger:
        :type task_trigger: TriggerTaskType
        :param machine_group_id:
        :type machine_group_id: int
        :param machine_group_trigger:
        :type machine_group_trigger: TriggerMachineGroupType
        :param observer_id:
        :type observer_id: str
        :param observer_trigger:
        :type observer_trigger: TriggerObserverType
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_events_serialize(
            trigger_type=trigger_type,
            task_id=task_id,
            task_trigger=task_trigger,
            machine_group_id=machine_group_id,
            machine_group_trigger=machine_group_trigger,
            observer_id=observer_id,
            observer_trigger=observer_trigger,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[EventInfo]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        ).data

    @validate_call
    def get_events_with_http_info(
        self,
        trigger_type: Optional[TriggerType] = None,
        task_id: Optional[StrictStr] = None,
        task_trigger: Optional[TriggerTaskType] = None,
        machine_group_id: Optional[StrictInt] = None,
        machine_group_trigger: Optional[TriggerMachineGroupType] = None,
        observer_id: Optional[StrictStr] = None,
        observer_trigger: Optional[TriggerObserverType] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> ApiResponse[List[EventInfo]]:
        """Get Events

        Fetch events based on filters.

        :param trigger_type:
        :type trigger_type: TriggerType
        :param task_id:
        :type task_id: str
        :param task_trigger:
        :type task_trigger: TriggerTaskType
        :param machine_group_id:
        :type machine_group_id: int
        :param machine_group_trigger:
        :type machine_group_trigger: TriggerMachineGroupType
        :param observer_id:
        :type observer_id: str
        :param observer_trigger:
        :type observer_trigger: TriggerObserverType
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_events_serialize(
            trigger_type=trigger_type,
            task_id=task_id,
            task_trigger=task_trigger,
            machine_group_id=machine_group_id,
            machine_group_trigger=machine_group_trigger,
            observer_id=observer_id,
            observer_trigger=observer_trigger,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[EventInfo]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        response_data.read()
        return self.api_client.response_deserialize(
            response_data=response_data,
            response_types_map=_response_types_map,
        )

    @validate_call
    def get_events_without_preload_content(
        self,
        trigger_type: Optional[TriggerType] = None,
        task_id: Optional[StrictStr] = None,
        task_trigger: Optional[TriggerTaskType] = None,
        machine_group_id: Optional[StrictInt] = None,
        machine_group_trigger: Optional[TriggerMachineGroupType] = None,
        observer_id: Optional[StrictStr] = None,
        observer_trigger: Optional[TriggerObserverType] = None,
        _request_timeout: Union[None, Annotated[StrictFloat,
                                                Field(gt=0)],
                                Tuple[Annotated[StrictFloat,
                                                Field(gt=0)],
                                      Annotated[StrictFloat,
                                                Field(gt=0)]]] = None,
        _request_auth: Optional[Dict[StrictStr, Any]] = None,
        _content_type: Optional[StrictStr] = None,
        _headers: Optional[Dict[StrictStr, Any]] = None,
        _host_index: Annotated[StrictInt, Field(ge=0, le=0)] = 0,
    ) -> RESTResponseType:
        """Get Events

        Fetch events based on filters.

        :param trigger_type:
        :type trigger_type: TriggerType
        :param task_id:
        :type task_id: str
        :param task_trigger:
        :type task_trigger: TriggerTaskType
        :param machine_group_id:
        :type machine_group_id: int
        :param machine_group_trigger:
        :type machine_group_trigger: TriggerMachineGroupType
        :param observer_id:
        :type observer_id: str
        :param observer_trigger:
        :type observer_trigger: TriggerObserverType
        :param _request_timeout: timeout setting for this request. If one
                                 number provided, it will be total request
                                 timeout. It can also be a pair (tuple) of
                                 (connection, read) timeouts.
        :type _request_timeout: int, tuple(int, int), optional
        :param _request_auth: set to override the auth_settings for an a single
                              request; this effectively ignores the
                              authentication in the spec for a single request.
        :type _request_auth: dict, optional
        :param _content_type: force content-type for the request.
        :type _content_type: str, Optional
        :param _headers: set to override the headers for a single
                         request; this effectively ignores the headers
                         in the spec for a single request.
        :type _headers: dict, optional
        :param _host_index: set to override the host_index for a single
                            request; this effectively ignores the host_index
                            in the spec for a single request.
        :type _host_index: int, optional
        :return: Returns the result object.
        """ # noqa: E501

        _param = self._get_events_serialize(
            trigger_type=trigger_type,
            task_id=task_id,
            task_trigger=task_trigger,
            machine_group_id=machine_group_id,
            machine_group_trigger=machine_group_trigger,
            observer_id=observer_id,
            observer_trigger=observer_trigger,
            _request_auth=_request_auth,
            _content_type=_content_type,
            _headers=_headers,
            _host_index=_host_index)

        _response_types_map: Dict[str, Optional[str]] = {
            '200': "List[EventInfo]",
            '422': "HTTPValidationError",
        }
        response_data = self.api_client.call_api(
            *_param, _request_timeout=_request_timeout)
        return response_data.response

    def _get_events_serialize(
        self,
        trigger_type,
        task_id,
        task_trigger,
        machine_group_id,
        machine_group_trigger,
        observer_id,
        observer_trigger,
        _request_auth,
        _content_type,
        _headers,
        _host_index,
    ) -> RequestSerialized:

        _host = None

        _collection_formats: Dict[str, str] = {}

        _path_params: Dict[str, str] = {}
        _query_params: List[Tuple[str, str]] = []
        _header_params: Dict[str, Optional[str]] = _headers or {}
        _form_params: List[Tuple[str, str]] = []
        _files: Dict[str, Union[str, bytes, List[str], List[bytes],
                                List[Tuple[str, bytes]]]] = {}
        _body_params: Optional[bytes] = None

        # process the path parameters
        # process the query parameters
        if trigger_type is not None:

            _query_params.append(('trigger_type', trigger_type.value))

        if task_id is not None:

            _query_params.append(('task_id', task_id))

        if task_trigger is not None:

            _query_params.append(('task_trigger', task_trigger.value))

        if machine_group_id is not None:

            _query_params.append(('machine_group_id', machine_group_id))

        if machine_group_trigger is not None:

            _query_params.append(
                ('machine_group_trigger', machine_group_trigger.value))

        if observer_id is not None:

            _query_params.append(('observer_id', observer_id))

        if observer_trigger is not None:

            _query_params.append(('observer_trigger', observer_trigger.value))

        # process the header parameters
        # process the form parameters
        # process the body parameter

        # set the HTTP header `Accept`
        if 'Accept' not in _header_params:
            _header_params['Accept'] = self.api_client.select_header_accept(
                ['application/json'])

        # authentication setting
        _auth_settings: List[str] = ['APIKeyHeader']

        return self.api_client.param_serialize(
            method='GET',
            resource_path='/events/',
            path_params=_path_params,
            query_params=_query_params,
            header_params=_header_params,
            body=_body_params,
            post_params=_form_params,
            files=_files,
            auth_settings=_auth_settings,
            collection_formats=_collection_formats,
            _host=_host,
            _request_auth=_request_auth)



"""List the tasks information via CLI."""
from typing import TextIO
import argparse
import sys

from inductiva import tasks, _cli, projects
from inductiva.utils import format_utils
from inductiva.client import models


def list_tasks(args: argparse.Namespace, fout: TextIO = sys.stdout):
    _list_tasks(**vars(args), fout=fout)


def _list_tasks(project_name, last_n, task_id, all_tasks: bool, fout: TextIO,
                **_):
    """ List the user's tasks. 

    Lists based on the flags (task_id, last_n or project_name).
    It can print the last N tasks (default value is 10)
    or a single task with a specific ID
    or all tasks of a project.
    """
    if project_name and task_id:
        print(
            "inductiva tasks list: error: "
            "argument -i/--id not allowed with argument -p/--project-name",
            file=sys.stderr)
        return 1

    last_n = -1 if all_tasks else last_n

    if task_id:
        task_list = [tasks.Task(i) for i in task_id]
    elif project_name:
        print(f"Showing tasks for project: {project_name}.", file=fout)
        task_list = projects.Project(project_name).get_tasks(last_n=last_n)
    else:
        task_list = tasks.get_tasks(last_n=last_n)

    if not task_list:
        print("No tasks found.", file=fout)
        return 1

    table_dict = tasks.to_dict(task_list)

    emph_formatter = format_utils.get_ansi_formatter()

    def color_formater(status):
        if status.upper() == models.TaskStatusCode.SUCCESS.upper():
            return emph_formatter(status, format_utils.Emphasis.GREEN)
        return status

    formatters = {
        "Submitted": [format_utils.datetime_formatter,],
        "Started": [format_utils.datetime_formatter,],
        "Status": [color_formater]
    }

    header_formatters = [
        lambda x: emph_formatter(x.upper(), format_utils.Emphasis.BOLD)
    ]

    print(format_utils.get_tabular_str(table_dict,
                                       formatters=formatters,
                                       header_formatters=header_formatters),
          file=fout,
          end="")
    print(
        "\nTo see more details about a task, "
        "use `inductiva tasks info <task_id>`.",)
    return 0


def register(parser):
    """Register the list user's tasks command."""

    subparser = parser.add_parser("list",
                                  help="List tasks.",
                                  formatter_class=argparse.RawTextHelpFormatter)

    subparser.description = ("The `inductiva tasks list` command provides "
                             "an overview of your tasks on the platform.\n"
                             "It lists the most recent tasks or a specific "
                             "task by ID.\n"
                             "You can control the number of tasks listed with "
                             "the '-n' or '--last-n' option.\n")

    _cli.utils.add_watch_argument(subparser)

    group = subparser.add_mutually_exclusive_group()
    group.add_argument("-n",
                       "--last-n",
                       type=int,
                       default=10,
                       help="List last N tasks. Default: 10.")
    group.add_argument("-a",
                       "--all",
                       dest="all_tasks",
                       action="store_true",
                       help="List all tasks.")
    group.add_argument("-i",
                       "--id",
                       type=str,
                       nargs="+",
                       default=None,
                       dest="task_id",
                       help="List a task with a specific ID.")
    #project e id da erro exclusivo
    subparser.add_argument("-p",
                           "--project-name",
                           type=str,
                           default=None,
                           help="List the tasks of a project.")

    subparser.set_defaults(func=list_tasks)



"""WRF simulator module of the API."""

from typing import List, Optional, Union

from inductiva import simulators, tasks, types
from inductiva.commands.commands import Command
from inductiva.commands.mpiconfig import MPIConfig


@simulators.simulator.mpi_enabled
class WRF(simulators.Simulator):
    """Class to invoke a generic WRF simulation on the API."""

    VALID_CASE_NAMES = [
        "em_b_wave",
        "em_convrad",
        "em_esmf_exp",
        "em_fire",
        "em_grav2d_x",
        "em_heldsuarez",
        "em_hill2d_x",
        "em_les",
        "em_quarter_ss",
        "em_real",
        "em_scm_xy",
        "em_seabreeze2d_x",
        "em_squall2d_x",
        "em_squall2d_y",
        "em_tropical_cyclone",
    ]

    def __init__(self, /, version: Optional[str] = None, use_dev: bool = False):
        """Initialize the WRF simulator.

        Args:
            version (str): The version of the simulator to use. If None, the
                latest available version in the platform is used.
            use_dev (bool): Request use of the development version of
                the simulator. By default (False), the production version
                is used.
        """
        super().__init__(version=version, use_dev=use_dev)
        self.simulator = "arbitrary_commands"
        self.simulator_name_alias = "wrf"

    def _validate_build_script(self, build_script: str) -> None:
        """
        Validates the build script by checking if required environment variables
        are correctly set.

        :param build_script: Path to the build script file.
        :raises ValueError: If any required variable is missing or incorrect.
        """

    def run(self,
            input_dir: Optional[str],
            *,
            on: types.ComputationalResources,
            use_hwthread: bool = True,
            case_name: str = "em_real",
            n_vcpus: Optional[int] = None,
            storage_dir: Optional[str] = "",
            resubmit_on_preemption: bool = False,
            remote_assets: Optional[Union[str, list[str]]] = None,
            init_commands: Optional[List[str]] = None,
            gen_gif: bool = False,
            gen_gif_files: Optional[List[str]] = None,
            gen_gif_output_dir: Optional[str] = ".",
            gen_gif_fps: int = 3,
            gen_gif_variable: Optional[List[str]] = "RAINNC",
            project: Optional[str] = None,
            time_to_live: Optional[str] = None,
            **kwargs) -> tasks.Task:
        """Run the simulation.

        Args:
            input_dir: Path to the directory of the simulation input files.
            on: The computational resource to launch the simulation on.
            n_vcpus: Number of vCPUs to use in the simulation. If not provided
                (default), all vCPUs will be used.
            use_hwthread: If specified Open MPI will attempt to discover the
                number of hardware threads on the node, and use that as the
                number of slots available.
            other arguments: See the documentation of the base class.
            resubmit_on_preemption (bool): Resubmit task for execution when
                previous execution attempts were preempted. Only applicable when
                using a preemptible resource, i.e., resource instantiated with
                `spot=True`.
            remote_assets: Additional remote files that will be copied to
                the simulation directory.
            init_commands: List of helper commands to prepare things for your
                simulation. Used to copy files from `/WRF`. It can also be used
                to run any helper function present within `/WRF/WPS`.
                Will run before the `wrf.exe` command.
            project: Name of the project to which the task will be
                assigned. If None, the task will be assigned to
                the default project. If the project does not exist, it will be
                created.
            gen_gif (bool): If True, generates a GIF from the simulation
                output files. Default is False.
            gen_gif_files (List[str]): List of ordered output files to be used
                for generating the GIF. Required if `gen_gif` is True.
                The order of the files will determine the order of frames in
                the GIF.
            gen_gif_output_dir (str): Directory where the GIF will be saved.
                Default is the current directory.
            gen_gif_fps (int): Frames per second for the GIF. Default is 3.
            gen_gif_variable (str): Variable to be used for generating the GIF.
                Default is "RAINNC".
            time_to_live: Maximum allowed runtime for the task, specified as a
                string duration. Supports common time duration formats such as
                "10m", "2 hours", "1h30m", or "90s". The task will be
                automatically terminated if it exceeds this duration after
                starting.
        """

        if case_name not in self.VALID_CASE_NAMES:
            raise ValueError(f"Invalid case name: {case_name}. "
                             f"Valid case names are: {self.VALID_CASE_NAMES}")

        if gen_gif and gen_gif_files is None:
            raise ValueError("If `gen_gif` is True, `gen_gif_files` must be "
                             "provided.")

        self._check_vcpus(n_vcpus, on)

        #only runs checks if we dont use remote assets

        self._input_files_exist(input_dir=input_dir,
                                remote_assets=remote_assets,
                                sim_config_filename="namelist.input")

        mpi_kwargs = {}
        mpi_kwargs["use_hwthread_cpus"] = use_hwthread
        if n_vcpus is not None:
            mpi_kwargs["np"] = n_vcpus
        mpi_config = MPIConfig(version="4.1.6", **mpi_kwargs)

        commands = [f"/scripts/create_links.sh /WRF/test/{case_name}"]

        if init_commands:
            commands += init_commands

        commands += [
            Command("./wrf.exe", mpi_config=mpi_config),
            "/scripts/delete_links.sh"
        ]

        if gen_gif:
            # if gen gif var is a single value turn it into a list
            if isinstance(gen_gif_variable, str):
                gen_gif_variable = [gen_gif_variable]
            for var in gen_gif_variable:
                files = " ".join(gen_gif_files)
                commands.append(
                    f"conda run -n wrf-env python /scripts/gen_gif.py "
                    f"--files {files} "
                    f"--output-dir {gen_gif_output_dir} "
                    f"--fps {gen_gif_fps} "
                    f"--var {var} ")

        return super().run(input_dir,
                           on=on,
                           commands=commands,
                           storage_dir=storage_dir,
                           remote_assets=remote_assets,
                           resubmit_on_preemption=resubmit_on_preemption,
                           project=project,
                           time_to_live=time_to_live,
                           **kwargs)



# coding: utf-8

"""
    InductivaWebAPI

    No description provided (generated by Openapi Generator https://github.com/openapitools/openapi-generator)

    The version of the OpenAPI document: 0.1.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501

from __future__ import annotations
import pprint
import re  # noqa: F401
import json

from datetime import datetime
from pydantic import BaseModel, ConfigDict, StrictBool, StrictInt, StrictStr
from typing import Any, ClassVar, Dict, List, Optional
from typing import Optional, Set
from typing_extensions import Self


class TaskRunnerRegisterInfo(BaseModel):
    """
    Info for creating a task-runner.
    """

  # noqa: E501
    machine_group_id: StrictStr
    create_time: datetime
    cpu_count_logical: StrictInt
    cpu_count_physical: StrictInt
    memory: StrictInt
    mpi_cluster: Optional[StrictBool] = False
    host_name: StrictStr
    host_id: StrictStr
    __properties: ClassVar[List[str]] = [
        "machine_group_id", "create_time", "cpu_count_logical",
        "cpu_count_physical", "memory", "mpi_cluster", "host_name", "host_id"
    ]

    model_config = ConfigDict(
        populate_by_name=True,
        validate_assignment=True,
        protected_namespaces=(),
    )

    def to_str(self) -> str:
        """Returns the string representation of the model using alias"""
        return pprint.pformat(self.model_dump(by_alias=True))

    def to_json(self) -> str:
        """Returns the JSON representation of the model using alias"""
        # TODO: pydantic v2: use .model_dump_json(by_alias=True, exclude_unset=True) instead
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_str: str) -> Optional[Self]:
        """Create an instance of TaskRunnerRegisterInfo from a JSON string"""
        return cls.from_dict(json.loads(json_str))

    def to_dict(self) -> Dict[str, Any]:
        """Return the dictionary representation of the model using alias.

        This has the following differences from calling pydantic's
        `self.model_dump(by_alias=True)`:

        * `None` is only added to the output dict for nullable fields that
          were set at model initialization. Other fields with value `None`
          are ignored.
        """
        excluded_fields: Set[str] = set([])

        _dict = self.model_dump(
            by_alias=True,
            exclude=excluded_fields,
            exclude_none=True,
        )
        return _dict

    @classmethod
    def from_dict(cls, obj: Optional[Dict[str, Any]]) -> Optional[Self]:
        """Create an instance of TaskRunnerRegisterInfo from a dict"""
        if obj is None:
            return None

        if not isinstance(obj, dict):
            return cls.model_validate(obj)

        _obj = cls.model_validate({
            "machine_group_id":
                obj.get("machine_group_id"),
            "create_time":
                obj.get("create_time"),
            "cpu_count_logical":
                obj.get("cpu_count_logical"),
            "cpu_count_physical":
                obj.get("cpu_count_physical"),
            "memory":
                obj.get("memory"),
            "mpi_cluster":
                obj.get("mpi_cluster")
                if obj.get("mpi_cluster") is not None else False,
            "host_name":
                obj.get("host_name"),
            "host_id":
                obj.get("host_id")
        })
        return _obj
